/*------------------------------------------------------------------------------
* gencorr.c : generate satellite clock corrections
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.12 $ $Date: 2007/03/20 08:49:29 $
* history : 2007/05/08 1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

#define C1          (FREQ1*FREQ1/(FREQ1*FREQ1-FREQ2*FREQ2))
#define C2          (FREQ2*FREQ2/(FREQ1*FREQ1-FREQ2*FREQ2))
#define MAXFILE     100        /* max number of input files */
#define TTOL        0.05       /* time tollerence (s) */
#define THRESSLIP   0.1        /* slip threshold of correction difference (m) */

static const char usage[]="gencorr [-i tint(s)][-c] obsfile [...] navfile";

/* print usage ---------------------------------------------------------------*/
static void printusage(void)
{
    fprintf(stderr,"%s\n",usage);
    exit(0);
}
/* epoch correction ----------------------------------------------------------*/
static void epcorr(const obsd_t *obs, int n, const double *rb,
                   nav_t *nav, int opt, double *corr)
{
    int i;
    double rs[MAXOBS][3],dts[MAXOBS],e[3],pr;
    gtime_t time;
    
    satpos(obs,n,nav,rs,dts,NULL);
    
    time=obs[0].time;
    for (i=0;i<n&&i<MAXOBS;i++) {
        if ((pr=geodist(rs[i],rb,e))>0.0) {
            if (!opt) {
                if (obs[i].L[0]!=0.0) corr[obs[i].sat-1]=lam[0]*obs[i].L[0]-pr;
            }
            else {
                if (obs[i].L[0]!=0.0&&obs[i].L[1]!=0.0)
                    corr[obs[i].sat-1]=C1*lam[0]*obs[i].L[0]-C2*lam[1]*obs[i].L[1]-pr;
            }
        }
    }
}
/* mean without 0.0 ----------------------------------------------------------*/
static double mean(const double *x, int n)
{
    int i,j=0;
    double sum=0.0;
    for (i=0;i<n;i++) if (x[i]!=0.0) {sum+=x[i]; j++;}
    return j<=0?0.0:sum/j;
}
/* generate correction -------------------------------------------------------*/
static void gencorr(const obsd_t *obs, int n, const nav_t *nav, int opt,
                    const double *rb, int cont)
{
    int i,j,k,m;
    char s[64];
    eph_t *e[MAXOBS],eph[MAXOBS];
    gtime_t time[31];
    double corr[31][MAXSAT]={{0}},dc[31][MAXSAT],t,tspan,mdc;
    nav_t nav0=*nav;
    nav0.eph=eph;
    
    /* epoch correction */
    for (i=j=m=0;i<n;i=j) {
        while (j<n&&timediff(obs[j].time,obs[i].time)<=0.0) j++;
        
        /* select ephemeris by first epoch */
        if (i==0) {
            seleph(obs[0].time,obs,j,nav,e);
            for (k=nav0.n=0;k<j;k++) if (e[i]) eph[nav0.n++]=*e[k];
        }
        time[m]=obs[i].time;
        epcorr(obs+i,j-i,rb,&nav0,opt,corr[m++]);
    }
    /* invalidate slip period (assume no receiver clock jump) */
    for (i=0;i<MAXSAT;i++) {
        for (j=0;j<m-1;j++) {
            if (fabs(corr[j][i]-corr[j+1][i])<=THRESSLIP) continue;
            for (j=0;j<m;j++) corr[j][i]=0.0;
            break;
        }
    }
    tspan=timediff(time[m-1],time[0]);
    
    /* substract interpolated correction */
    for (i=0;i<MAXSAT;i++) {
        for (j=0;j<m;j++) {
            t=timediff(time[j],time[0]);
            dc[j][i]=corr[j][i]-corr[0][i]-(corr[m-1][i]-corr[0][i])*t/tspan;
        }
    }
    /* output corrections */
    for (i=0;i<(cont?m-1:m);i++) {
        mdc=mean(dc[i],MAXSAT);
        for (j=0;j<MAXSAT;j++) if (dc[i][j]!=0.0) dc[i][j]-=mdc;
        
        time2str(time[i],s,0);
        printf("%s",s);
        for (j=0;j<MAXSAT;j++) {
            printf(" %5.1f%s",dc[i][j]*1E3,j==MAXSAT-1?"\n":"");
        }
    }
}
/* gencorr main ---------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int i,j,k,nfile=0,opt=0;
    char *files[MAXFILE];
    obs_t obs;
    nav_t nav;
    double tt,tint=30.0;
    double rb[]={-3855263.0423,3427432.5685,3741020.3095}; /* USUD */
    
    for (i=1;i<argc;i++) {
        if      (!strcmp(argv[i],"-i")&&i+1<argc) tint=atof(argv[++i]);
        else if (!strcmp(argv[i],"-r")&&i+3<argc)
            for (j=0;j<3;j++) rb[j]=atof(argv[++i]);
        else if (!strcmp(argv[i],"-c")) opt=1;
        else if (*argv[i]=='-') printusage();
        else if (nfile<MAXFILE) files[nfile++]=argv[i];
    }
    if (readrnx(files,nfile,&obs,&nav)<=0) {
        fprintf(stderr,"no observation data.\n");
        return -1;
    }
    if (nav.n<=0) {
        fprintf(stderr,"no navgation message.\n");
        return -1;
    }
    for (i=0;i<obs.n;i=k) {
        while (fmod(time2gpst(obs.data[i].time,NULL)+TTOL,tint)>TTOL*2.0) i++;
        for (j=k=i;j<obs.n;j++) {
            tt=timediff(obs.data[j].time,obs.data[i].time);
            if (tt<tint-TTOL) k++; else if (tt>tint+TTOL) break;
        }
        gencorr(obs.data+i,j-i,&nav,opt,rb,k<obs.n);
    }
    return 0;
}
