/*------------------------------------------------------------------------------
* pos2kml.c : convert positions to google earth kml file
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* options : -DLIBRTKP = libary mode
*
* version : $Revision: 3 $ $Date: 07/05/09 14:13 $
* history : 2007/01/20  1.0 new
*           2007/03/15  1.1 modify color sequence
*           2007/04/03  1.2 add geodetic height option
*                           support input of NMEA GGA sentence
*                           delete altitude info for track
*                           add time stamp option
*                           separate readsol.c file
*-----------------------------------------------------------------------------*/
#include <stdarg.h>
#include "rtklib.h"

static const char rcsid[]="$Id: pos2kml.c,v 1.3 2007/03/29 19:38:18 TTAKA Exp TTAKA $";

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" usage: pos2kml [option]... file [...]",
"",
" Read position file(s) and convert it to Google Earth KML file. Each line in",
" the input file shall contain fields of time, position fields (latitude/",
" longitude/height or x/y/z-ecef), and quality flag(option). The line started",
" with '%', '#', ';' is treated as comment. Command options are as follows.",
" ([]:default)",
"",
" -h        print help",
" -o file   output file [infile + .kml]",
" -c color  track color (0:off,1:white,2:green,3:orange,4:red,5:yellow) [5]",
" -p color  point color (0:off,1:white,2:green,3:orange,4:red,5:by qflag) [5]",
" -a        output altitude information [off]",
" -ag       output geodetic altitude [off]",
" -tg       output time stamp of gpst [off]",
" -tu       output time stamp of utc [gpst]",
" -i tint   output time interval (s) (0:all) [0]",
" -q qflg   output q-flags (0:all) [0]",
" -f n e h  add north/east/height offset to position (m) [0 0 0]",
" -e        input x/y/z-ecef position [latitude/longitude/height]",
" -n        input NMEA-0183 GGA sentence [off]",
" -g        input latitude/longitude in the form of ddd mm ss.ss' [ddd.ddd]",
" -s sep    field separator [' ']"
};

#define POSMAX   100000         /* max number of position records */
#define SIZP     0.2            /* mark size of rover positions */
#define SIZR     0.4            /* mark size of reference position */

/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* print message -------------------------------------------------------------*/
#ifdef LIBRTKP
extern void printmsg(const char *format, ...);
#else
static void printmsg(const char *format, ...)
{
    va_list arg;
    va_start(arg,format); vfprintf(stderr,format,arg); va_end(arg);
    fprintf(stderr,"\n");
}
#endif /* LIBRTKP */

/* output track --------------------------------------------------------------*/
static void outtrack(FILE *f, double sol[][NSOL], int n, const double mpos[],
                     const char *color, int outalt)
{
    int i;
    double alt=0.0,p[2];
    
    fprintf(f,"<Placemark>\n");
    fprintf(f,"  <name>Rover Track</name>\n");
    fprintf(f,"  <LookAt>\n");
    fprintf(f,"    <longitude>%14.9f</longitude>\n",mpos[1]);
    fprintf(f,"    <latitude>%14.9f</latitude>\n",mpos[0]);
    fprintf(f,"  </LookAt>\n");
    fprintf(f,"  <Style>\n");
    fprintf(f,"    <LineStyle>\n");
    fprintf(f,"      <color>%s</color>\n",color);
    fprintf(f,"    </LineStyle>\n");
    fprintf(f,"  </Style>\n");
    fprintf(f,"  <LineString>\n");
    if (outalt) fprintf(f,"    <altitudeMode>absolute</altitudeMode>\n");
    fprintf(f,"    <coordinates>\n");
    for (i=0;i<n;i++) {
        if (outalt) {
            p[0]=sol[i][0]*D2R; p[1]=sol[i][1]*D2R;
            alt=sol[i][2]-(outalt==2?geoidh(p):0.0);
        }
        fprintf(f,"      %14.9f,%14.9f,%8.3f\n",sol[i][1],sol[i][0],alt);
    }
    fprintf(f,"    </coordinates>\n");
    fprintf(f,"  </LineString>\n");
    fprintf(f,"</Placemark>\n");
}
/* output point --------------------------------------------------------------*/
static void outpoint(FILE *f, gtime_t time, const double sol[], const char *label,
                     double size, int color, int outalt, int outtime)
{
    char *icon,*url="http://maps.google.com/mapfiles/kml/";
    double alt=0.0,p[2],ep[6];

    switch (color) {
        case 1:  icon="pal2/icon18.png"; break; /* white */
        case 2:  icon="pal3/icon53.png"; break; /* green */
        case 3:  icon="pal3/icon49.png"; break; /* orange */
        default: icon="pal4/icon55.png"; break; /* red */
    }
    fprintf(f,"<Placemark>\n");
    if (*label) fprintf(f,"  <name>%s</name>\n",label);
    if (outtime==1) time=gpst2utc(time);
    if (outtime) {
        time2epoch(time,ep);
        fprintf(f,"  <TimeStamp><when>%04.0f-%02.0f-%02.0fT%02.0f:%02.0f:%05.2fZ</when></TimeStamp>\n",
                ep[0],ep[1],ep[2],ep[3],ep[4],ep[5]);
    }
    fprintf(f,"  <Style>\n");
    fprintf(f,"    <IconStyle>\n");
    fprintf(f,"      <scale>%.1f</scale>\n",size);
    fprintf(f,"      <Icon><href>%s%s</href></Icon>\n",url,icon);
    fprintf(f,"    </IconStyle>\n");
    fprintf(f,"  </Style>\n");
    fprintf(f,"  <Point>\n");
    if (outalt) {
        fprintf(f,"    <extrude>1</extrude>\n");
        fprintf(f,"    <altitudeMode>absolute</altitudeMode>\n");
        p[0]=sol[0]*D2R; p[1]=sol[1]*D2R;
        alt=sol[2]-(outalt==2?geoidh(p):0.0);
    }
    fprintf(f,"    <coordinates>%14.9f,%14.9f,%8.3f</coordinates>\n",sol[1],
            sol[0],alt);
    fprintf(f,"  </Point>\n");
    fprintf(f,"</Placemark>\n");
}
/* save kml file -------------------------------------------------------------*/
static void savekml(const char *file, const gtime_t time[], double sol[][NSOL],
                    int n, const double rpos[], const double mpos[], int tcolor,
                    int pcolor, int outalt, int outtime)
{
    FILE *f;
    int i;
    char *color[]={"ffffffff","ff00ff00","ff00aaff","ff0000ff","ff00ffff"};
    
    f=fopen(file,"w");
    if (f<0) {
        fprintf(stderr,"file open error : %s\n",file);
        return;
    }
    fprintf(f,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    fprintf(f,"<kml xmlns=\"http://earth.google.com/kml/2.0\">\n");
    fprintf(f,"<Document>\n");
    if (tcolor>0) {
        outtrack(f,sol,n,mpos,color[tcolor-1],outalt);
    }
    if (pcolor>0) {
        fprintf(f,"<Folder>\n");
        fprintf(f,"  <name>Rover Position</name>\n");
        for (i=0;i<n;i++) {
            outpoint(f,time[i],sol[i],"",SIZP,pcolor==5?(int)sol[i][3]+1:pcolor,
                     outalt,outtime);
        }
        fprintf(f,"</Folder>\n");
    }
    if (rpos[1]!=0.0&&rpos[3]>=-100.0) {
        outpoint(f,time[0],rpos,"Reference Position",SIZR,1,outalt,0);
    }
    fprintf(f,"</Document>\n");
    fprintf(f,"</kml>\n");
    fclose(f);
}
/* convert position file to google earth kml file ----------------------------*/
extern void pos2kml(const char *infile, const char *outfile, gtime_t ts,
                    gtime_t te, double *offset, int tcolor, int pcolor,
                    int posf, int degf, int outalt, int outtime, double tint,
                    int qflg, const char *sep)
{
    static double sol[POSMAX][NSOL];
    static gtime_t time[POSMAX];
    int i,j,k,n,q,nq[6]={0};
    char *p,file[256],msg[256];
    double rpos[3]={0.0},mpos[3]={0.0};
    
    if (!*outfile) {
        if ((p=strrchr(infile,'.'))) {
            strncpy(file,infile,p-infile);
            strcpy(file+(p-infile),".kml");
        }
        else sprintf(file,"%s.kml",infile);
    }
    else strcpy(file,outfile);
    
    /* read solution file */
    if ((n=readsol(infile,posf,degf,tint,qflg,sep,POSMAX,time,sol,rpos))<=0) {
        printmsg("no position data : %s",infile);
        return;
    }
    if (timediff(ts,te)<0.0) {
        for (i=j=0;i<n;i++) {
            if (timediff(time[i],ts)<0.0) continue;
            if (timediff(time[i],te)>0.0) continue;
            time[j]=time[i];
            for (k=0;k<4;k++) sol[j][k]=sol[i][k];
            j++;
        }
        n=j;
    }
    for (i=0;i<n;i++) if ((q=(int)sol[i][3])>0) nq[q<6?q-1:5]++;
    
    /* mean position */
    for (i=0;i<3;i++) {
        for (j=0;j<n;j++) mpos[i]+=sol[j][i];
        mpos[i]/=n;
    }
    /* add offset */
    offset[0]/=RE*D2R;
    offset[1]/=RE*D2R*cos(mpos[0]*D2R);
    for (i=0;i<n;i++) {
        for (j=0;j<3;j++) sol[i][j]+=offset[j];
    }
    if (rpos[0]!=0.0) {
        for (j=0;j<3;j++) rpos[j]+=offset[j];
    }
    /* save kml file */
    savekml(file,time,sol,n,rpos,mpos,tcolor,pcolor,outalt,outtime);
                                    
    /* print counts */
    sprintf(msg,"n=%d Q=",n);
    for (i=1;i<=5;i++) {
        if (nq[i-1]<=0) continue;
        sprintf(msg+strlen(msg),"%d:%d(%.1f%%) ",i,nq[i-1],n<=0?0.0:100.0*nq[i-1]/n);
    }
    printmsg("converted : %s",msg);
}
/* pos2kml main --------------------------------------------------------------*/
#ifndef LIBRTKP
int main(int argc, char **argv)
{
    int i,j,n,posf=0,degf=0,outalt=0,outtime=0,qflg=0,tcolor=5,pcolor=5;
    char *infile[32],*outfile="",*sep=" ";
    double offset[3]={0.0},tint=0.0,es[6]={2000,1,1},ee[6]={2000,1,1};
    gtime_t ts={0},te={0};
    
    for (i=1,n=0;i<argc;i++) {
        if      (!strcmp(argv[i],"-o")&&i+1<argc) outfile=argv[++i];
        else if (!strcmp(argv[i],"-ts")&&i+2<argc) {
            sscanf(argv[++i],"%lf/%lf/%lf",es  ,es+1,es+2);
            sscanf(argv[++i],"%lf:%lf:%lf",es+3,es+4,es+5);
            ts=epoch2time(es);
        }
        else if (!strcmp(argv[i],"-te")&&i+2<argc) {
            sscanf(argv[++i],"%lf/%lf/%lf",ee  ,ee+1,ee+2);
            sscanf(argv[++i],"%lf:%lf:%lf",ee+3,ee+4,ee+5);
            te=epoch2time(ee);
        }
        else if (!strcmp(argv[i],"-c")&&i+1<argc) tcolor=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-p")&&i+1<argc) pcolor=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-s")&&i+1<argc) sep=argv[++i];
        else if (!strcmp(argv[i],"-f")&&i+3<argc) {
            for (j=0;j<3;j++) offset[j]=atof(argv[++i]);
        }
        else if (!strcmp(argv[i],"-e")) posf=1;
        else if (!strcmp(argv[i],"-n")) posf=2;
        else if (!strcmp(argv[i],"-g")) degf=1;
        else if (!strcmp(argv[i],"-a")) outalt=1;
        else if (!strcmp(argv[i],"-ag")) outalt=2;
        else if (!strcmp(argv[i],"-tg")) outtime=1;
        else if (!strcmp(argv[i],"-tu")) outtime=2;
        else if (!strcmp(argv[i],"-i")&&i+i<argc) tint=atof(argv[++i]);
        else if (!strcmp(argv[i],"-q")&&i+i<argc) qflg=atoi(argv[++i]);
        else if (*argv[i]=='-') printhelp();
        else if (n<32) infile[n++]=argv[i];
    }
    if (tcolor<0||5<tcolor||pcolor<0||5<pcolor) {
        printmsg("%s : command option error",__FILE__);
        return -1;
    }
    if (n<=0) {
        printmsg("%s : no input file",__FILE__);
        return -1;
    }
    for (i=0;i<n;i++) {
        pos2kml(infile[i],outfile,ts,te,offset,tcolor,pcolor,posf,degf,outalt,
                outtime,tint,qflg,sep);
    }
    return 0;
}
#endif /* LIBRTKP */
