/*------------------------------------------------------------------------------
* readsol.c : read position solution file
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1 $ $Date: 07/05/09 14:09 $
* history : 2007/04/14  1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

#define MAXFIELD    32          /* max number of fields in a record */
#define MAXNMEA    128          /* max length of nmea sentence */

/* fields to numbers ---------------------------------------------------------*/
static int tonum(char *p, const char *sep, double v[])
{
    char *q;
    int n=0,m=strlen(sep);
    
    if (!strcmp(sep," ")) {
        for (p=strtok(p,sep);p&&n<MAXFIELD;p=strtok(NULL,sep)) v[n++]=atof(p);
        return n;
    }
    while (n<MAXFIELD) {
        if ((q=strstr(p,sep))) *q='\0'; 
        v[n++]=atof(p);
        if (q) p=q+m; else break;
    }
    return n;
}
/* dd/mm/ss to deg ----------------------------------------------------------*/
static double dms2deg(const double dms[])
{
    return dms[0]+dms[1]/60.0+dms[2]/3600.0;
}
/* ddmm.mm to deg ------------------------------------------------------------*/
static double dmm2deg(double dmm)
{
    return floor(dmm/100.0)+fmod(dmm,100.0)/60.0;
}
/* separate time field -------------------------------------------------------*/
static void septime(double t, double *t1, double *t2, double *t3)
{
    *t1=floor(t/10000.0); t-=*t1*10000.0; *t2=floor(t/100.0); *t3=t-*t2*100.0;
}
/* date/time to time ---------------------------------------------------------*/
static gtime_t datet2time(double date, double time)
{
    double ep[6];
    septime(date,ep+2,ep+1,ep);
    septime(time,ep+3,ep+4,ep+5);
    ep[0]+=ep[0]<80.0?2000.0:1900.0;
    return epoch2time(ep);
}
/* screen solution by time and q-flag ---------------------------------------*/
static int screensol(gtime_t t, double sol[], double tint, int qflg)
{
    double tt=fmod(time2gpst(t,NULL),tint);
    return (tint<=0.0||(0.01<tt&&tt<tint-0.01))&&(qflg<=0||((int)sol[3]==qflg));
}
/* decode nmea sentence -----------------------------------------------------*/
static int decodenmea(const char *sent, gtime_t *time, double sol[])
{
    static double date=10100.0;
    static int qs[]={0,5,4,0,2,1};
    double t,lat,lon,vel,dir,hdop,alt=0.0,msl=0.0;
    int q,nrcv;
    char ns,ew;
    if (!strncmp(sent,"$GPRMC",6)) {
        if (sscanf(sent+7,"%lf,A,%lf,%c,%lf,%c,%lf,%lf,%lf",
                   &t,&lat,&ns,&lon,&ew,&vel,&dir,&date)<8) return -1;
        return -1;
    }
    else if (!strncmp(sent,"$GPGGA",6)) {
        if (sscanf(sent+7,"%lf,%lf,%c,%lf,%c,%d,%d,%lf,%lf,M,%lf,M",
                   &t,&lat,&ns,&lon,&ew,&q,&nrcv,&hdop,&alt,&msl)<10) return -1;
        sol[3]=qs[q<0?0:(q<6?q:0)];
    }
    else return -1;
    *time=datet2time(date,t);
    sol[0]=(ns=='N'?1.0:-1.0)*dmm2deg(lat);
    sol[1]=(ew=='E'?1.0:-1.0)*dmm2deg(lon);
    sol[2]=alt+msl;
    return 0;
}
/* read nmea file -----------------------------------------------------------*/
static int readnmea(FILE *fp, double tint, int qflg, int nmax, gtime_t time[],
                    double sol[][NSOL])
{
    int i,j=-1,n=0,nb;
    char buff[2048],sent[MAXNMEA+1];
    
    while ((nb=fread(buff,1,sizeof(buff),fp))>0) {
        for (i=0;i<nb;i++) {
            if (buff[i]=='$') j=0;
            if (j<0) continue;
            if ((sent[j++]=buff[i])=='\n') {
                sent[j]='\0';
                if (!decodenmea(sent,&time[n],sol[n])&&
                    screensol(time[n],sol[n],tint,qflg)) n++;
                j=-1;
            }
            if ((j==3&&strncmp(sent,"$GP",3))||j>=MAXNMEA) j=-1;
        }
    }
    return n;
}
/* decode position record ----------------------------------------------------*/
static int decodepos(char *buff, int posf, int degf, const char *sep,
                     gtime_t *time, double sol[], double rpos[], int *week)
{
    int i,m;
    char *p;
    double v[MAXFIELD];
    
    if (*buff=='%'||*buff=='#'||*buff==';') { /* comment line */
        if (strstr(buff,"ref pos")&&(p=strchr(buff,':'))) {
            m=tonum(p+1,sep,v);
            if (posf==1&&m>=3) {
                ecef2pos(v,rpos); rpos[0]*=R2D; rpos[1]*=R2D;
            }
            else if (degf&&m>=7) {
                rpos[0]=dms2deg(v); rpos[1]=dms2deg(v+3); rpos[2]=v[6];
            }
            else if (m>=3) {
                for (i=0;i<3;i++) rpos[i]=v[i];
            }
        }
        else if (strstr(buff,"obs start")&&(p=strstr(buff,"gpsweek"))) {
            sscanf(p,"gpsweek%d",week);
        }
        return -1;
    }
    if (sscanf(buff,"%lf/%lf/%lf %lf:%lf:%lf",v,v+1,v+2,v+3,v+4,v+5)>=6) {
        *time=epoch2time(v); p=buff+11;
    }
    else if (sscanf(buff,"%lf",v)>=1) {
        *time=gpst2time(*week,v[0]); p=buff;
    }
    else return -1;
    m=tonum(p,sep,v);
    
    if (posf==1&&m>=4) { /* x/y/z-ecef */
        ecef2pos(v+1,sol); sol[0]*=R2D; sol[1]*=R2D; sol[3]=m>4?v[4]:0.0;
    }
    else if (degf&&m>=8) { /* lat/lon/height (ddd mm ss) */
        sol[0]=dms2deg(v+1); sol[1]=dms2deg(v+4);
        sol[2]=v[7]; sol[3]=m>8?v[8]:0.0;
    }
    else if (m>=4) { /* lat/lon/height (ddd.dddd) */
        for (i=0;i<4;i++) sol[i]=v[i+1]; sol[3]=m>4?v[4]:0.0;
    }
    return m>=4?0:-1;
}
/* read position file -------------------------------------------------------*/
static int readpos(FILE *fp, int posf, int degf, double tint, int qflg,
                   const char *sep, int nmax, gtime_t time[],
                   double sol[][NSOL], double rpos[])
{
    int n=0,week=1400;
    char buff[256];
    while (n<nmax&&fgets(buff,sizeof(buff),fp)) {
        if (!decodepos(buff,posf,degf,sep,&time[n],sol[n],rpos,&week)&&
            screensol(time[n],sol[n],tint,qflg)) n++;
    }
    return n;
}
/* read solution file ---------------------------------------------------------
* read position solution file
* args   : char   *file     I  position solution file
*          int    posf      I  solution form (0:lat,lon,hgt,1:x,y,z-ecef,2:nmea)
*          int    degf      I  latitude/longitude form (0:ddd.ddd,1:ddd mm ss)
*          double tint      I  input time interval (sec) (0.0:all)
*          int    qflg      I  input quality flag (0:all)
*          char   *sep      I  field separator ("":space)
*          int    nmax      I  max number of solutions
*          gtime_t time[]   O  time of solutions
*          double sol[][]   O  solutions {lat,lon,height,q-flag}
*          double rpos[]    O  reference station position {lat,lon,height}
* return : number of solutions
* note   : pos[i] = time[i] solution lat/lon(deg),height(m),quality-flag
*-----------------------------------------------------------------------------*/
extern int readsol(const char *file, int posf, int degf, double tint, int qflg,
                   const char *sep, int nmax, gtime_t time[],
                   double sol[][NSOL], double rpos[])
{
    FILE *fp;
    int n;
    if (!(fp=fopen(file,"rb"))) return 0;
    if (posf==2) n=readnmea(fp,tint,qflg,nmax,time,sol);
    else n=readpos(fp,posf,degf,tint,qflg,sep,nmax,time,sol,rpos);
    fclose(fp);
    return n;
}
